set FlowgenDebug 0

Class Flow

Flow instproc init { name } {
	$self instvar started_ name_
	set started_ false
	set name_ $name
}

Flow instproc setFlowSet { fs } {
	$self instvar fs_
	set fs_ $fs
}

Flow instproc setup { _nodes fid cid protocol traffictype traffic_args } {
	$self instvar srcAgent_ dstAgent_ protocol_ tg_ name_ ttype_ fileSize_
	global seqnum

	set src [lindex $_nodes 0]
	set dst [lindex $_nodes 1]
	set protocol_ $protocol
	set ttype_ $traffictype
	switch $protocol_ {
		UDP {
			set srcAgent_ [new Agent/UDP]
			set dstAgent_ [new Agent/UDP]
		}
		TCP {
			set srcAgent_ [new Agent/TCP/Newreno]
			if { false } {
				set cwndtracefile [open ~/tmp/out/cwnd$fid.tr w]
				$srcAgent_ trace t_seqno_
				$srcAgent_ trace ack_
				$srcAgent_ trace dupacks_
				$srcAgent_ trace cwnd_
				$srcAgent_ trace credit_
				$srcAgent_ trace ssthresh_
				$srcAgent_ attach $cwndtracefile
			}
			set dstAgent_ [new Agent/TCPSink]
		}
		TCPX {
			set srcAgent_ [new Agent/TCP/Newreno]
			$srcAgent_ set credit_enable_ 1
			if { true } {
				set cwndtracefile [open ~/tmp/out/cwnd$fid.tr w]
				$srcAgent_ trace t_seqno_
				$srcAgent_ trace ack_
				$srcAgent_ trace dupacks_
				$srcAgent_ trace cwnd_
				$srcAgent_ trace credit_
				$srcAgent_ trace ssthresh_
				$srcAgent_ attach $cwndtracefile
			}
			set dstAgent_ [new Agent/TCPSink]
		}
		TCPSack {
			set srcAgent_ [new Agent/TCP/Sack1]
			set dstAgent_ [new Agent/TCPSink/Sack1]
		}
		TFRC {
			set srcAgent_ [new Agent/TFRC]
			set dstAgent_ [new Agent/TFRCSink]
		}
		default {
			puts " $name_: Flow/setup error - unknown protocol $protocol"
			exit 1
		}
	}
	$srcAgent_ set fid_ $fid
	$srcAgent_ set prio_ $cid
	set ns [Simulator instance]
	$ns attach-agent $src $srcAgent_
	$ns attach-agent $dst $dstAgent_
	$ns connect $srcAgent_ $dstAgent_

	global FlowgenDebug
	if { $FlowgenDebug == 1 } { puts " $name_: creating fid/cid $fid/$cid" }

	switch $traffictype {
		CBR {
			$srcAgent_ set packetSize_ [lindex $traffic_args 0]
			set tg_ [new Application/Traffic/$traffictype]
			$tg_ set packetSize_ [lindex $traffic_args 0]
			$tg_ set rate_       [lindex $traffic_args 1]
			$tg_ set random_     [lindex $traffic_args 2]  
		}
		Poisson {
			$srcAgent_ set packetSize_ [lindex $traffic_args 0]
			set tg_ [new Application/Traffic/$traffictype]
			$tg_ set packetSize_ [lindex $traffic_args 0]
			$tg_ set linkrate_   [lindex $traffic_args 1]
			$tg_ set rate_       [lindex $traffic_args 2]
			set myRNG_ [new RNG]
			for {set i 1} {$i <= $seqnum} {incr i} { $myRNG_ next-substream }
			$tg_ use-rng $myRNG_
		}
		Trace {
			set tg_ [new Application/Traffic/$traffictype]
			$tg_ attach-tracefile [lindex $traffic_args 0]
		}
		Pareto {
			$srcAgent_ set packetSize_ [lindex $traffic_args 0]
			set tg_ [new Application/Traffic/$traffictype]
			$tg_ set packetSize_ [lindex $traffic_args 0]
			$tg_ set rate_       [lindex $traffic_args 1]
			$tg_ set burst_time_ [lindex $traffic_args 2]
			$tg_ set idle_time_  [lindex $traffic_args 3]   
			$tg_ set shape_      [lindex $traffic_args 4]
			set myRNG_ [new RNG]
			for {set i 1} {$i <= $seqnum} {incr i} { $myRNG_ next-substream }
			$tg_ use-rng $myRNG_
		}
		Greedy {
			$srcAgent_ set packetSize_ [lindex $traffic_args 0]
			set tg_ [new Application/FTP]
			$tg_ set packetSize_ [lindex $traffic_args 0]
			$srcAgent_ set overhead_ [lindex $traffic_args 1]
		}
		File {
			$srcAgent_ set packetSize_ [lindex $traffic_args 0]
			set tg_ [new Application/FTP]
			$tg_ set packetSize_ [lindex $traffic_args 0]
			$srcAgent_ set overhead_ [lindex $traffic_args 1]
			set fileSize_ [lindex $traffic_args 2]
			$srcAgent_ proc done {} "$self finish-file"
		}
		Trace {
			set tg_ [new Application/Traffic/$traffictype]
			$tg_ attach-tracefile [lindex $traffic_args 0]
		}
		default {
			puts " $name_: Flow/setup error - unsupported traffic type $traffictype"
			exit 1
		}
	}
	$tg_ attach-agent $srcAgent_
}

Flow instproc start {} {
	$self instvar tg_ started_ srcAgent_ name_ protocol_ ttype_
	global ns
	set started_ true
	global FlowgenDebug
	if { $FlowgenDebug == 1 } { puts " $ttype_ $name_: starting fid/cid [$srcAgent_ set fid_]/[$srcAgent_ set prio_] at [format "%02.4f" [[Simulator instance] now]]" }
	if { $ttype_ == "File" } { $self start-file
	} elseif { $protocol_ != "TFRC"} { $tg_ start } else { $srcAgent_ start }
}

Flow instproc stop {} {
	$self instvar tg_ started_ srcAgent_ name_ protocol_ ttype_ 
	if { $started_ == "false" } { return }
	set started_ false
	global FlowgenDebug
	if { $FlowgenDebug == 1 } { puts " $name_: stopping fid/cid [$srcAgent_ set fid_]/[$srcAgent_ set prio_] at [format "%02.4f" [[Simulator instance] now]]" }
	if { $protocol_ != "TFRC"} { $tg_ stop } else { $srcAgent_ stop }
}

Flow instproc start-file {} {
	$self instvar tg_ fileSize_ srcAgent_ stime_
	global ns
	set stime_ [$ns now]
	$tg_ producemore $fileSize_
}

Flow instproc finish-file {} {
	$self instvar tg_ started_ srcAgent_ name_ fs_ stime_ tracefile_
	global ns
	if [info exists tracefile_ ] {
		if { $started_ == "true" } {
			set now [$ns now]
			puts $tracefile_ [format "flow%i %f %f" [$srcAgent_ set fid_] [expr $now - $stime_] $now] }
	}
	set started_ false
	global FlowgenDebug
	if { $FlowgenDebug == 1 } { puts " $name_: finishing fid/cid [$srcAgent_ set fid_]/[$srcAgent_ set prio_] at [format "%02.4f" [[Simulator instance] now]]" }
	if [info exists fs_] { $fs_ flow-done }
}

Flow instproc attach { tf } {
	$self instvar tracefile_
	set tracefile_ $tf
}

Class FlowSet

FlowSet instproc init { name } {
	$self instvar name_
	set name_ $name
}

FlowSet instproc setup { _nodes fid fcnt cid protocol traffictype traffic_args } {
	$self instvar flows_ fcnt_ lastflow_ name_
	
	set fcnt_ $fcnt
	set lastflow_ -1
	for {set i 0} {$i < $fcnt_} {incr i} {
		set flows_($i) [new Flow "$name_ [expr $fid + $i]"]
		if { $cid < 0 } {
			set cc [expr $fid + $i]
		} else {
			set cc $cid
		}
		$flows_($i) setup $_nodes $fid $cc $protocol $traffictype $traffic_args
		$flows_($i) setFlowSet $self
	}
}

FlowSet instproc setFlowArrival { iat_avg iat_mode fd_avg fd_mode } {
	$self instvar fd_src_ fd_avg_ iat_src_ iat_avg_ name_
	set iat_avg_ $iat_avg
	set fd_avg_ $fd_avg
	switch $iat_mode {
		exp {
			set iat_src_ [new RandomVariable/Exponential]
			$iat_src_ set avg_ $iat_avg
		}
		fixed {
			set iat_src_ [new RandomVariable/Constant]
			$iat_src_ set val_ $iat_avg
		}
		uni {
			set iat_src_ [new RandomVariable/Uniform]
			$iat_src_ set min_ [expr 0.01 * $iat_avg] 
			$iat_src_ set max_ [expr 1.99 * $iat_avg]
		}
		default {
			puts " $name_: FlowSet/setFlowArrival error - unknown flow arrival mode $iat_mode"
			exit 1
		}
	}
	switch $fd_mode {
		exp {
			set fd_src_ [new RandomVariable/Exponential]
			$fd_src_ set avg_ $fd_avg
		}
		fixed {
			set fd_src_ [new RandomVariable/Constant]
			$fd_src_ set val_ $fd_avg
		}
		uni {
			set fd_src_ [new RandomVariable/Uniform]
			$fd_src_ set min_ [expr 0.0 * $fd_avg] 
			$fd_src_ set max_ [expr 2.0 * $fd_avg]
		}
		default {
			puts " $name_: FlowSet/setFlowArrival error - unknown flow duration mode $fd_mode"
			exit 1
		}
	}
}

FlowSet instproc start {} {
	$self instvar lastflow_ name_ wrapup_
	if { $lastflow_ != -1 } {
		puts " $name_: FlowSet/start error - flowset is already started, stop first"
		exit 1
	}
	set wrapup_ 0
	$self startNextFlow
}

FlowSet instproc stop {} {
	$self instvar flows_ fcnt_ lastflow_ wrapup_
	for {set i 0} {$i < $fcnt_} {incr i} {$flows_($i) stop}
	set wrapup_ 1
	set lastflow_ -1
}

FlowSet instproc flow-done {} {
	$self instvar iat_avg_
	if { $iat_avg_ == 0 } { $self startNextFlow }
}

# internal method
FlowSet instproc startNextFlow { } {
	$self instvar flows_ fcnt_ lastflow_ fd_src_ fd_avg_ iat_src_ iat_avg_ name_ wrapup_
	
	if { $wrapup_ == 1 } { return }
	set i [expr ($lastflow_ + 1) % $fcnt_]
	while {0 == 0} {
		if { [$flows_($i) set started_] == "false" } {
			if { $fd_avg_ != 0 } {
				[Simulator instance] after [$fd_src_ value] "$flows_($i) stop"
			}
			set lastflow_ $i
			$flows_($i) start
	break
		} else {
			set i [expr ($i + 1) % $fcnt_]
		}
		if { $i == $lastflow_ } {
			global FlowgenDebug
			if { $FlowgenDebug == 1 } { puts "\n$name_: cannot start flow - all busy at [format "%02.4f" [[Simulator instance] now]]" }
	break
		}
	}
# TODO: shoud check here whether flow has been started and do proper action
	if { $iat_avg_ != 0 } {
		set nextflowstart [ expr [[Simulator instance] now] + [$iat_src_ value] ]
		[Simulator instance] at $nextflowstart "$self startNextFlow"
	}
}

FlowSet instproc attach { tf } {
	$self instvar fcnt_ flows_
	for {set i 0} {$i < $fcnt_} {incr i} {$flows_($i) attach $tf}
}
