# flow monitoring: arrival rate, departure rate, drop rate, max delay, mean delay, drop count

proc bytes-to-rate { bytes time } {
	return [expr $bytes*8.0/$time]
}

# dump the flow into file 
#   link - monitoring link
#   fm   - flow monitor
#   interval - dumping interval
#   flowTrace - file channel
proc dump-flow {link fm interval flowTrace} {
	global ns dsf
	$ns at [expr [$ns now] + $interval] "dump-flow $link $fm $interval $flowTrace" 

	foreach f [$fm flows] {
		# <flowname> <time> <send bytes> <recv bytes>
		# <max delay> <sum delay> <pkts>

		set delaysample [$f get-delay-samples]
		set delayavg 0.0
		if { [$delaysample cnt] > 0 } { set delayavg [$delaysample mean] }

		puts $flowTrace [format " %3if %.2ft %10u %10u %8.6f %8.6f %6u %6u %6u" \
			[$f set flowid_] [$ns now] [$f set barrivals_] [$f set bdepartures_] \
			[$delaysample max] $delayavg [$f set parrivals_] [$f set pdepartures_] \
			[$f set pdrops_]]
		$delaysample reset
		$f reset
	}
}

proc flow-mon { blink flowMonInterval flowTrace} {
	global ns opt
	set fmon [$ns makeflowmon Fid]
	$ns attach-fmon $blink $fmon
	$ns at 1.0 "dump-flow $blink $fmon $flowMonInterval $flowTrace"
}
