#ifndef _DSFBASE_H_
#define _DSFBASE_H_ 1

struct Pkt {
  DSFPACKET p;
  PacketTime deadline;
  double ts;
  Pkt(DSFPACKET p, PacketTime d, double t) : p(p), deadline(d), ts(t) {}
};

struct ServiceClass {
  PacketTime targetDelay;
  list<Pkt> pktQ;
  int service;
  int backlog;
  int credit;
  ServiceClass() : targetDelay(-1), service(0), backlog(0), credit(0) {}
};

struct Slot {
  int cnum;
  int len;
  Slot(int c, int l) : cnum(c), len(l) {}
};

struct Segment {
  PacketTime capacity;
  PacketTime occupancy;
  list<Slot> slots;
  Segment() : capacity(0), occupancy(0) {}
};

class DSFBase {
  // configuration
public:
  static const int codeCount = 32;
  bool traceflag;
private:
  int segmentCount;
  int defaultClass;
  int queueLimit; // FIFO mode
  bool prioMode;  // PRIO mode
  bool abeMode;   // ABE mode
  bool ldMode;    // LD mode
  int vsize;      // used in ABE mode

  // scheduling
  ServiceClass svcClass[codeCount];
  Segment segment[codeCount];
  int codeMap[codeCount];
  PacketTime currentTime;
  int backlog;

  inline void trace( DSFQ caller, char op, int cnum, int plen, double ts = -1 );
public:
  DSFBase() : traceflag(false), segmentCount(0), defaultClass(0), queueLimit(0),
    prioMode(false), abeMode(false), ldMode(false),
    vsize(0), currentTime(0), backlog(0) {}
  void setQueueLimit(int ql) { queueLimit = ql; }
  void addClass(int cnum, PacketTime td) {
    svcClass[cnum].targetDelay = td;
//    fprintf(stderr, "add class %d %lu\n", cnum, td);
  }
  inline void setLD();
  inline void setABE();
  inline void setPRIO();
  inline void reset();
  inline int enqueue( DSFPACKET p, int cnum, DSFQ caller );
  inline DSFPACKET dequeue( double& delay, DSFQ caller  );
};

#endif /* _DSFBASE_H_ */
