#ifndef _DSFBASE2_H_
#define _DSFBASE2_H_ 1

void DSFBase::setLD() {
  DSFASSERT(!prioMode && !abeMode && !ldMode);
  DSFASSERT(queueLimit);
  ldMode = true;
}

void DSFBase::setABE() {
  DSFASSERT(!abeMode && !prioMode && !ldMode);
  DSFASSERT(!queueLimit);
  abeMode = true;
}

void DSFBase::setPRIO() {
  DSFASSERT(!prioMode && !abeMode && !ldMode);
  DSFASSERT(queueLimit);
  prioMode = true;
}

void DSFBase::trace( DSFQ caller, char op, int cnum, int plen, double ts ) {
  if (!traceflag) return;
  char buf[512];
  int len = sprintf( buf, "%c %4u %6i %10.6f %4zu %6i %6i %6i",
    op, cnum, plen, DSFNOW, svcClass[cnum].pktQ.size(),
    svcClass[cnum].service, svcClass[cnum].backlog, backlog );
  if ( ts != -1 ) len += sprintf( buf + len, " %10.6f", ts );
  DSFTRACE(caller, buf);
}

void DSFBase::reset() {
  multimap<PacketTime,int> delays;
  for (int i = 0; i < codeCount; i += 1) {
    if (svcClass[i].targetDelay != -1) {
      delays.insert(pair<PacketTime,int>(svcClass[i].targetDelay, i));
    }
  }
  segmentCount = delays.size();
                                                   DSFASSERT(!delays.empty());
  PacketTime prevClassDelay = 0;
  for (int i = 0; i < segmentCount; i += 1) {
//    fprintf( stderr, "DSF: creating segment with capacity %ld bytes, code point %i\n", delays.begin()->first - prevClassDelay, delays.begin()->second );
    segment[i].capacity = delays.begin()->first - prevClassDelay;
    prevClassDelay = delays.begin()->first;
    defaultClass = delays.begin()->second;
    codeMap[defaultClass] = i;
    delays.erase(delays.begin());
  }
}

int DSFBase::enqueue( DSFPACKET p, int cnum, DSFQ caller ) {
                                   DSFASSERT(cnum >= 0 && cnum < segmentCount);
  int plen = DSFPLEN(p);
  bool dsfMode = !queueLimit;
  ServiceClass& sc = (dsfMode || prioMode || ldMode) ? svcClass[cnum] : svcClass[0];
                                                DSFASSERT(sc.targetDelay >= 0);

  // Simplified ABE: always accept Green
  if (abeMode) {                                      // ABE mode
    if (vsize < svcClass[1].targetDelay) {
      vsize += plen;
    } else if (cnum != 0) {             // accept all green packets
      trace( caller, 'D', cnum, plen );               // *D*rop
      DSFDROP(caller, p);
      return 0;
    }
    sc.pktQ.push_back( Pkt(p, currentTime + sc.targetDelay, DSFNOW) );
    sc.backlog += plen;
    backlog += plen;
    trace( caller, 'A', cnum, plen );                 // *A*cceptance
    return 1;
  }

  int limit = queueLimit;
  int* testbacklog = &backlog;

  if (dsfMode) {                                      // DSF mode
    for (int sn = 0; sn <= codeMap[cnum]; sn += 1) {
      if (segment[sn].occupancy < segment[sn].capacity) {
        segment[sn].slots.push_back(Slot(cnum, plen));
        segment[sn].occupancy += plen;
        sc.service += plen;
        trace( caller, 'E', cnum, plen );             // slot *E*ntry into segment
    break;
      }
    }
    limit = sc.service;
    testbacklog = &sc.backlog;
  } else if (ldMode) {
    if (segment[0].occupancy < queueLimit) {
      segment[0].slots.push_back(Slot(cnum, plen));
      segment[0].occupancy += plen;
      sc.service += plen;
      trace( caller, 'E', cnum, plen );             // slot *E*ntry into segment
    }
    limit = sc.service;
    testbacklog = &sc.backlog;
  }

//  if (dsfMode)                                       // drop-front only for DSF (not FIFO, etc)
  while (*testbacklog && *testbacklog >= limit) {
    if (prioMode && sc.pktQ.empty()) break;
                                                   DSFASSERT(!sc.pktQ.empty());
    double ts = sc.pktQ.front().ts;                   // drop front
    DSFPACKET p = sc.pktQ.front().p;
    sc.pktQ.pop_front();
    int droplen = DSFPLEN(p);
    sc.backlog -= droplen;
    backlog -= droplen;
    trace( caller, 'P', cnum, droplen, DSFNOW - ts ); // *P*ushout
    DSFDROP(caller, p);
  }

  if (*testbacklog < limit) {
    sc.pktQ.push_back( Pkt(p, currentTime + sc.targetDelay, DSFNOW) );
    sc.backlog += plen;
    backlog += plen;
    trace( caller, 'A', cnum, plen );                 // *A*cceptance
    return 1;
  } else {
    trace( caller, 'D', cnum, plen );                 // *D*rop
    DSFDROP(caller, p);
    return 0;
  }
}

/* The front slot in the slot segment might cover multiple smaller packets
 * and is thus not removed until all these packets are serviced.  */
DSFPACKET DSFBase::dequeue( double& delay, DSFQ caller ) {

  // Simplified ABE: always serve Green, if Blue has room -> set g = 1
  if (abeMode) {                                      // ABE mode
    int sizeGreen = svcClass[0].pktQ.empty() ? 0 : DSFPLEN(svcClass[0].pktQ.front().p);
    bool canServeGreen = svcClass[1].pktQ.empty() ||
      currentTime + sizeGreen < svcClass[1].pktQ.front().deadline;
    int startClass = canServeGreen ? 0 : 1;
    for (int cnum = startClass; cnum < 2; cnum += 1) {
      ServiceClass& sc = svcClass[cnum];
      while (!sc.pktQ.empty()) {
        delay = DSFNOW - sc.pktQ.front().ts;
        DSFPACKET p = sc.pktQ.front().p;            // get next packet
        PacketTime deadline = sc.pktQ.front().deadline;
        sc.pktQ.pop_front();
        int plen = DSFPLEN(p);
        sc.backlog -= plen;
        backlog -= plen;
        if (cnum == 1 || currentTime < deadline) {  // still early enough?
          currentTime += plen;
          vsize -= plen;
          if (vsize < 0) vsize = 0;
          trace( caller, 'S', cnum, plen, delay );  // *S*ervice
          return p;
        } else {
          trace( caller, 'L', cnum, plen, delay );  // *L*ate drop
          DSFLDROP(caller, p);
        }
      }
      DSFASSERT(sc.backlog == 0);
    }
    DSFASSERT(backlog == 0);
    currentTime = 0;
    vsize = 0;
    return NULL;
  }

  if (queueLimit > 0 && !ldMode) {                    // FIFO mode
    for (int c = 0; c < segmentCount; c += 1) {
      ServiceClass& sc = svcClass[c];
      if (!sc.pktQ.empty()) {
        double ts = sc.pktQ.front().ts;
        DSFPACKET p = sc.pktQ.front().p;
        sc.pktQ.pop_front();
        int plen = DSFPLEN(p);
        sc.backlog -= plen;
        backlog -= plen;
        delay = DSFNOW - ts;
        trace( caller, 'S', c, plen, delay );         // *S*ervice
        return p;
      }
      DSFASSERT(sc.backlog == 0);
    }
    DSFASSERT(backlog == 0);
    return NULL;
  }

  for (int sn = 0; sn < segmentCount; sn += 1) {      // DSF, LD mode
    while (!segment[sn].slots.empty()) {
      int cnum = segment[sn].slots.front().cnum;
      int slen = segment[sn].slots.front().len;
      segment[sn].slots.pop_front();
      segment[sn].occupancy -= slen;
      ServiceClass& sc = svcClass[cnum];              // slot -> service class
      sc.service -= slen;
      sc.credit += slen;
      if (sc.credit > 0) {
        while (!sc.pktQ.empty()) {
          delay = DSFNOW - sc.pktQ.front().ts;
          DSFPACKET p = sc.pktQ.front().p;            // get next packet
          PacketTime deadline = sc.pktQ.front().deadline;
          sc.pktQ.pop_front();
          int plen = DSFPLEN(p);
          sc.backlog -= plen;
          backlog -= plen;
          if (currentTime < deadline) {               // still early enough?
            currentTime += plen;
            sc.credit -= plen;
            if (sc.credit > 0) {                      // service leftover?
              segment[sn].slots.push_front(Slot(cnum,0));
            }
            trace( caller, 'S', cnum, plen, delay );  // *S*ervice
            return p;
          } else {
            trace( caller, 'L', cnum, plen, delay );  // *L*ate drop
            DSFLDROP(caller, p);
          }
        }
        trace( caller, 'U', cnum, sc.credit );        // *U*nused credit
        sc.credit = 0;
      }
    }
  }

  // slot queue runs empty -> reset system
  for (int i = 0; i < codeCount; i += 1) {
                                          DSFASSERT(segment[i].occupancy == 0);
    ServiceClass& sc = svcClass[i];
#if 0 // should not be needed
    while (!sc.pktQ.empty()) {
      double d = DSFNOW - sc.pktQ.front().ts;
      DSFPACKET p = sc.pktQ.front().p;
      sc.pktQ.pop_front();
      int plen = DSFPLEN(p);
      sc.backlog -= plen;
      backlog -= plen;
      trace( caller, 'C', i, plen, d );              // *C*leanup
      DSFDROP(caller, p);
    }
#endif
                                          DSFASSERT(sc.pktQ.empty());
                                          DSFASSERT(sc.service >= 0);
                                          DSFASSERT(sc.backlog == 0);
                                          DSFASSERT(sc.credit <= 0);
    sc.service = 0;
    sc.credit = 0;
    DSFRESET(caller);
  }
                                          DSFASSERT(backlog == 0);
  currentTime = 0;
  return NULL;
}

#endif /* _DSFBASE2_H_ */
