#include <stdlib.h>
 
#include "random.h"
#include "trafgen.h"
#include "ranvar.h"

class PoissonTraffic : public TrafficGenerator {
 public:
	PoissonTraffic();
	virtual double next_interval(int&);
	int command(int argc, const char*const* argv);
 protected:
	virtual void start();
	void init();
	double rate_;     /* send rate during on time (bps) */
	double linkrate_;     /* send rate during on time (bps) */
	int seqno_;
	int maxpkts_;
	ExponentialRandomVariable factor;
};


static class PoissonTrafficClass : public TclClass {
 public:
	PoissonTrafficClass() : TclClass("Application/Traffic/Poisson") {}
	TclObject* create(int, const char*const*) {
		return (new PoissonTraffic());
	}
} class_cbr_traffic;

PoissonTraffic::PoissonTraffic() : seqno_(0), factor(0.0)
{
	bind_bw("rate_", &rate_);
	bind_bw("linkrate_", &linkrate_);
	bind("packetSize_", &size_);
	bind("maxpkts_", &maxpkts_);
}

void PoissonTraffic::init()
{
  factor.setavg(1.0);
	if (agent_)
		if (agent_->get_pkttype() != PT_TCP && agent_->get_pkttype() != PT_TFRC)
			agent_->set_pkttype(PT_CBR);
}

void PoissonTraffic::start()
{
        init();
        running_ = 1;
        timeout();
}

double PoissonTraffic::next_interval(int& size)
{
	// Recompute interval in case rate_ or size_ has changes
	double t = factor.value() * (double)(size_ << 3) / (double)rate_;
//	while (t < (double)(size_ << 3) / (double)linkrate_) {
//	  t = t + factor.value() * (double)(size_ << 3) / (double)rate_;
//	}
	size = size_;
	if (++seqno_ < maxpkts_)
		return(t);
	else
		return(-1); 
}

int PoissonTraffic::command(int argc, const char*const* argv) {
  if (argc==3) {
    if (strcmp(argv[1], "use-rng") == 0) {
      factor.seed((char *)argv[2]);
      return (TCL_OK);
    }
  }
  return TrafficGenerator::command(argc,argv);
}
