#include "queue.h"                        // ns-2
#include "delay.h"                            // ns-2
#include <list>
using namespace std;

class InstVar {
public:
  static double bw_atof(const char* s);
  static double time_atof(const char* s);
};

typedef long long PacketTime;

class SPRI_Queue : public Queue {
  int linkRate;
	PacketTime currentTime;
	int backlog;
	int queueLimit;

	static const int fc = 16;
	int fb[fc];
	list<Packet*> pq[fc];

	Tcl_Channel tchan_;
public:
  SPRI_Queue() : linkRate(0) {}
protected:
	void reset() {
		currentTime = 0;
		backlog = 0;
		for (int i = 0; i < fc; i += 1) {
			fb[i] = 0;
			pq[i].clear();
		}
	}
	void enque(Packet* p) {
		int fid = HDR_IP(p)->prio() % fc;
		if (backlog < queueLimit) {
			backlog += HDR_CMN(p)->size();
			fb[fid] += HDR_CMN(p)->size();
			HDR_CMN(p)->timestamp() = Scheduler::instance().clock();
			pq[fid].push_back(p);
			trace( 'A', fid, Scheduler::instance().clock() );
		} else {
			trace( 'D', fid, Scheduler::instance().clock() );
			drop(p);
		}
	}
	Packet* deque() {
	  int fid;
		for (fid = 0; fid < fc; fid += 1) {
			if (!pq[fid].empty()) goto packetfound;
		}
		return NULL;
packetfound:
		Packet* p = pq[fid].front();
		pq[fid].pop_front();
		backlog -= HDR_CMN(p)->size();
		fb[fid] -= HDR_CMN(p)->size();
		currentTime += HDR_CMN(p)->size();
		trace( 'S', fid, HDR_CMN(p)->timestamp() );
		return p;
	}
	virtual int command( int argc, const char*const* argv ) {
		Tcl& tcl = Tcl::instance();
		switch (argc) {
		case 3:
		if (strcmp(argv[1], "attach") == 0) {
			int mode;
			tchan_ = Tcl_GetChannel( tcl.interp(), (char*)argv[2], &mode );
			if ( !tchan_ ) {
				tcl.resultf( "SPRI: can't attach trace %s for writing", argv[2] );
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "target") == 0) {
			LinkDelay* del = (LinkDelay*)TclObject::lookup( argv[2] );
			if ( del == 0 ) {
				tcl.resultf( "SPRI: no LinkDelay object %s", argv[2] );
				return (TCL_ERROR);
			}
			// second "target" call comes from flow monitor attachment (BW 0)
			if ( !linkRate ) linkRate = del->bandwidth() / 8;
			return Queue::command( argc, argv );
		}
		if (strcmp(argv[1], "set-queue-limit") == 0) {
		  if ( !linkRate ) {
		    tcl.resultf( "SPRI: link rate not set" );
		    return (TCL_ERROR);
		  }
		  queueLimit = linkRate * InstVar::time_atof(argv[2]);
//      fprintf( stderr, "SPRI: queueLimit %d\n", queueLimit);
		  return (TCL_OK);
		}
		}
		return Queue::command(argc, argv);
	}
	virtual void trace( char op, int prio, double ts ) {
		if (!tchan_) return;
		char wrk[512];
		sprintf( wrk, "P %c %4u %10.6f %14llu %6u %6u %10.6f\n",
			op, prio, Scheduler::instance().clock(), currentTime,
			backlog, fb[prio], Scheduler::instance().clock() - ts
		);
		Tcl_Write( tchan_, wrk, strlen(wrk) );
	}
};

// hook into Tcl for object creation
static class SPRIClass : public TclClass {
public:
	SPRIClass() : TclClass("Queue/SPRI") {}
	TclObject* create(int, const char*const*) { return new SPRI_Queue; }
} class_spri;
